/*
 * NanoWidgets v0.1
 * (C) 1999 Screen Media AS
 * 
 * Written by Vidar Hokstad
 * 
 * Contains code from The Nano Toolkit,
 * (C) 1999 by Alexander Peuchert.
 *
 */

#include "nclass.h"
#include "nwidget.h"
#include "nbutton.h"
#include "nslider.h"
#include "ntoggle.h"
#include "nradio.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>


/********* Text field *******************************************/

DEFINE_NOBJECT(textfield,widget)
    char * textbuf;

    long maxsize;    /* Maximum number of characters */
    int curpos;      /* Current cursor position */
    int firstpos;    /* First visible character */
    long overwrite;  /* Overwrite if 1, insert if 0 */

    int esc;         /* 1 if currently processing an escape sequence */

    /* Handler to be called to verify input - Can be used to restrict input
     * to for instance integers, or hexadecimal, or whatever format you want
     */
    int (* verify_handler)(struct textfield_nobject *,char *);
END_NOBJECT


DEFINE_NCLASS(textfield,widget)
    NSLOT(int,init);
    NSLOT(void,settext);
    NSLOT(const char * ,gettext);
END_NCLASS

#define n_textfield_init(__this__,__parent__,__text__) n_call(textfield,init,__this__,(__this__,__parent__,__text__))

typedef struct textfield_nobject NTEXTFIELD;


/* PRIVATE */
static void textfield_drawcursor (NTEXTFIELD * this, int h)
{
   int th,tw,tb;
   if (h == 0) n_widget_getgeometry(this,0,0,0,&h);   
   n_widget_setfg(this,RED);
   if (this->curpos > this->firstpos) n_widget_textextent(this,this->textbuf + this->firstpos,this->curpos - this->firstpos,&tw,&th,&tb);
   else tw = 0;
   /* n_widget_setmode(this,GR_MODE_XOR); */
   n_widget_line(this,3+tw,3,3+tw,h-4);
   /* n_widget_setmode(this,GR_MODE_SET); */
}

/* PRIVATE */
/* Temporary since GR_MODE_XOR above doesnt seem to work */
static void textfield_clearcursor (NTEXTFIELD * this, int h)
{
   int th,tw,tb;
   if (h == 0) n_widget_getgeometry(this,0,0,0,&h);   
   n_widget_setfg(this,CYAN);
   if (this->curpos > this->firstpos) n_widget_textextent(this,this->textbuf + this->firstpos,this->curpos - this->firstpos,&tw,&th,&tb);
   else tw = 0;
   /* n_widget_setmode(this,GR_MODE_XOR); */
   n_widget_line(this,3+tw,3,3+tw,h-4);
   /* n_widget_setmode(this,GR_MODE_SET); */
}


static void textfield_repaint (NTEXTFIELD * this)
{
   int w,h;
   int th,tw,tb,len;
   
   n_widget_getgeometry(this,0,0,&w,&h);
   n_widget_setfg(this,BLACK);
   n_widget_setbg(this,BLACK);
   n_widget_rect(this,0,0,w,h);
   
   n_widget_setfg(this,CYAN);
   n_widget_setbg(this,CYAN);
   n_widget_fillrect(this,2,2,w-4,h-4);

   n_widget_setfg(this,BLUE);

   len = strlen(this->textbuf + this->firstpos);

   tw = th = tb = 0;
   n_widget_textextent(this,this->textbuf + this->firstpos,len,&tw,&th,&tb);

   if (tw > w) printf("tw > w: tw = %d, w = %d\n",tw,w);
   /* FIXME: Ugly and slow */
   while (tw > (w-8)) {
      len--;
      n_widget_textextent(this,this->textbuf + this->firstpos,len,&tw,&th,&tb);
   }
   n_widget_text(this,4,h - 2 - tb,this->textbuf+this->firstpos,len);

   if (n_widget_isinfocus(this)) textfield_drawcursor(this,h);
}


static void textfield_settext (NTEXTFIELD * this, const char * text)
{   
   int maxsize;

   if (text && strlen(text) > 256) maxsize = strlen(text) + 1;
   else maxsize = 256;

   if (this->maxsize < maxsize) {
      if (this->textbuf) free(this->textbuf);
      if (this->textbuf = malloc(maxsize)) {
	 this->textbuf[0] = '\0';
	 this->maxsize = maxsize;
      } else this->maxsize = 0;
   }
   
   if (!this->textbuf) free(this->textbuf);
   if (text) this->textbuf = strdup(text);
   else this->textbuf = 0;

   if (text && this->textbuf) strcpy(this->textbuf,text);
   this->curpos = 0;
   this->firstpos = 0;
   
   n_widget_repaint(this);

}


static void textfield_cleanup (NTEXTFIELD * this)
{

   if (this->textbuf) free(this->textbuf);
   n_super(object,cleanup,this,(this));
}


static int textfield_init (NTEXTFIELD * this, NWIDGET * parent, const char * text)
{
   
   if (n_widget_init(this,parent)) return -1;

   /* FIXME: Should use n_textfield_settext */
   textfield_settext(this,text);
}


static void textfield_buttondown(NTEXTFIELD * this, int x, int y, BUTTON b)
{
   int len;
   int tw,th,tb;
   
   if (!this->textbuf) this->curpos = 0;
   else {
      /* FIXME: Do real calculation... This is *UGLY* */
      this->curpos = this->firstpos;
      tw = 0;
      while(tw < x) {
	 n_widget_textextent(this,this->textbuf + this->firstpos,this->curpos + 1 - this->firstpos,&tw,&th,&tb);
	 this->curpos++;
      }
      this->curpos -= 1;      
      
      len = strlen(this->textbuf);
      if (this->curpos > len) this->curpos = len;
   }
   n_widget_repaint(this);
}

/* PRIVATE */
/* Adjust this->firstpos so that this->curpos is always visible
 * FIXME: Currently this entire function is one *HUGE* hack. Should really
 * cache the last visible position, and use that, and recalc that when
 * firstpos is changed.
 */
static int textfield_adjustfirstpos(NTEXTFIELD * this)
{
   int w,h;
   int th,tw,tb,len;
   
   if (this->curpos <= this->firstpos) {
      this->curpos = this->firstpos;
      return 0;
   }
   
   n_widget_getgeometry(this,0,0,&w,&h);

   len = this->curpos - this->firstpos;

   tw = th = tb = 0;
   n_widget_textextent(this,this->textbuf + this->firstpos,len,&tw,&th,&tb);

   if (tw < (w-8)) return 0;
   while (len>0 && tw > (w-8)) {
      this->firstpos++;
      len--;
      n_widget_textextent(this,this->textbuf + this->firstpos,len,&tw,&th,&tb);
   }
   return 1;
}


/* FIXME: Clean this up... Logic is way too convoluted.
 * This will also need to be improved to handle Unicode text in the future.
 */
static void textfield_keypress(NTEXTFIELD * this, GR_CHAR ch, MODIFIER modifiers, BUTTON buttons)
{
   fprintf(stderr,"   ch = %c / %x\n   modifiers = %x\n",ch,ch,modifiers);
   if (this->textbuf && this->curpos < this->maxsize) {
      /* Erase the cursor if it is tere... */
      if (n_widget_isinfocus(this)) textfield_clearcursor(this,0);
      
      /* FIXME: Check modifiers */
      /* FIXME: Handle insert in addition to overwrite */
      if (this->esc) {
	 if (ch == 'D') {
	    if (!this->curpos) {
	       this->esc = 0;
	       if (n_widget_isinfocus(this)) textfield_drawcursor(this,0);
	       return;
	    }
	    this->curpos--;
	    this->esc = 0;
	    if (this->curpos < this->firstpos && this->firstpos) this->firstpos--;
	    else {
	       if (n_widget_isinfocus(this)) textfield_drawcursor(this,0);
	       return;
	    }
	 } else if (ch == 'C') {
	    if (this->textbuf && this->textbuf[this->curpos]) this->curpos++;
	    this->esc = 0;
	    
	    if (textfield_adjustfirstpos(this)) n_widget_repaint(this);
	    else if (n_widget_isinfocus(this)) textfield_drawcursor(this,0);
	    return;
	 } else {
	    if (n_widget_isinfocus(this)) textfield_drawcursor(this,0);
	    return;
	 }
      } else if (ch == 0x01) {  /* CTRL-A  -- Go to beginning */
	 this->curpos = 0;
	 this->firstpos = 0;

      } else if (ch == 0x05) {  /* CTRL-E  -- Go to end */
	 this->curpos = strlen(this->textbuf);
	 
	 if (textfield_adjustfirstpos(this)) n_widget_repaint(this);
	 else if (n_widget_isinfocus(this)) textfield_drawcursor(this,0);
	 return;
      } else if (ch == 0x18) {  /* CTRL-X  -- Delete line */
	 if (this->textbuf) this->textbuf[0] = '\0';
	 this->curpos = 0;
	 this->firstpos = 0;
      } else if (ch == 0x1b) {
	 this->esc = 1;
	 if (n_widget_isinfocus(this)) textfield_drawcursor(this,0);
	 return;
      } else if (ch == 0x7f) {
	 if (this->textbuf && this->curpos>0) {
	    memmove(this->textbuf + this->curpos-1, this->textbuf + this->curpos, this->maxsize - this->curpos);
	    this->curpos--;
	    if (this->curpos < this->firstpos) this->firstpos--;
	 }
      } else if (ch >= 0x20) {
	 if (this->overwrite) {
	    if (this->textbuf[this->curpos] == '\0') this->textbuf[this->curpos+1] = '\0';
	 } else {
	    memmove(this->textbuf + this->curpos + 1, this->textbuf + this->curpos, this->maxsize - this->curpos+1);
	 }
	 this->textbuf[this->curpos] = ch;
	 this->curpos++;
	 textfield_adjustfirstpos(this);
      }
      /* FIXME: Use verify handler */
      /* FIXME: check if we must scroll to keep cursor visible */
      n_widget_repaint(this);
   }
}

INIT_NCLASS(textfield,widget)
   NMETHOD(textfield,init,textfield_init);
   NMETHOD(object,cleanup,textfield_cleanup);
   NMETHOD(widget,repaint,textfield_repaint);
   NMETHOD(widget,buttondown,textfield_buttondown);
   NMETHOD(widget,keypress,textfield_keypress);
END_NCLASS


/****************** Test ****************************************/


void onclick(NBUTTON * w, BUTTON b)
{
   printf("Button %d was clicked in widget %p\n",b,w);
}

int main ()
{
   NWIDGET * w;
   NWIDGET * w2;
   
   NBUTTON * b1;
   NBUTTON * b2;
   NBUTTON * b3;
   NBUTTON * b4;

   NSLIDER * s;

   NRADIO * r1;
   NRADIO * r2;

   NTEXTFIELD * t;
   
   n_init_button_class();
   n_init_slider_class();
   n_init_toggle_class();
   n_init_radio_class();
   n_init_textfield_class();
   
   /* run demo.sh it sleeps instead*/
   /*sleep(2);*/

   w = NEW_NOBJECT(widget);
   w2 = NEW_NOBJECT(widget);

   b1 = NEW_NOBJECT(button);
   b2 = NEW_NOBJECT(button);
   b3 = NEW_NOBJECT(button);
   b4 = NEW_NOBJECT(toggle);
   
   s = NEW_NOBJECT(slider);
   
   r1 = NEW_NOBJECT(radio);
   r2 = NEW_NOBJECT(radio);
   
   t = NEW_NOBJECT(textfield);
   
   fprintf(stderr,"w = %p\nw2 = %p\nb = %p\n",w,w2,b1);
   
   n_widget_init(w,0);   
   n_widget_resize(w,300,200);
   n_widget_show(w);
   
   n_widget_init(w2,0);   
   n_widget_move(w2,300,300);
   n_widget_resize(w2,400,300);
   n_widget_show(w2);

   n_button_init(b1,w,"text1");
   n_widget_resize(b1,30,30);
   n_button_onclick(b1,onclick);
   n_widget_show(b1);
   
   n_button_init(b2,w2,"test");
   n_widget_resize(b2,30,30);
   n_button_onclick(b2,onclick);
   n_widget_show(b2);

   n_button_init(b3,w,"label");
   n_widget_move(b3,10,10);
   n_widget_resize(b3,60,30);
   n_button_onclick(b3,onclick);
   n_widget_show(b3);
   
   n_toggle_init(b4,w);
   n_widget_move(b4,10,90);
   n_widget_resize(b4,15,15);
   n_widget_show(b4);

   n_slider_init(s,w);
   n_widget_move(s,200,100);
   n_widget_resize(s,100,100);
   n_widget_show(s);
   
   n_radio_init(r1,w2);
   n_widget_move(r1,20,20);
   n_widget_resize(r1,20,20);
   n_widget_show(r1);
   
   n_radio_init(r2,w2);
   n_widget_move(r2,45,20);
   n_widget_resize(r2,20,20);
   n_radio_connect(r1,r2);
   n_widget_show(r2);

   
   n_textfield_init(t,w2,"This is my text");
   n_widget_move(t,45,100);
   n_widget_resize(t,200,20);
   n_widget_show(t);

   while (1) n_handle_event();
   
   n_widget_hide(w);
   n_object_cleanup(w);
   
   GrClose();
}
